from __future__ import annotations

from typing import Any
import click
from prettytable import PrettyTable

import ckan.plugins.toolkit as tk


@click.argument("module", required=True)
def update_module_info(module: str):    
    tk.get_action("update_last_module_check")({
        "ignore_auth": True},
        {"module": module,
        }
    )


def get_selfinfo():
    data = tk.get_action("get_selfinfo")({
        "ignore_auth": True},
        {}
    )

    platform_info: dict[str, Any] = data.get('platform_info', {})
    ram_usage: dict[str, Any] = data.get('ram_usage', {})
    platform_table = PrettyTable(["System", "Python Version", "RAM Usage %", "RAM Usage GB"])
    
    platform_table.add_row([
        platform_info.get("platform"),
        platform_info.get("python_version"),
        ram_usage.get("precent_usage"),
        ram_usage.get("used_ram"),
        ]
    )
    click.echo(click.style("System Information", fg='green'))
    
    click.echo(platform_table)
    groups_order: list[dict[str, Any]] = [
        {"name": "ckan", "label": "CKAN"},
        {"name": "ckanext", "label": "CKANEXT"},
        {"name": "other", "label": "Other"},
    ]
    groups:dict[str, Any] = data.get('groups', {})
    
    for item in groups_order:
        click.echo(click.style(item["label"], fg='green'))
        item_table = PrettyTable(["Name", "Current Version", "Possible Version", "Last Checked"])

        for key, values in groups.get(item["name"], {}).items():
            name = values.get('name')
            if name:
                item_table.add_row([
                    name,
                    values.get("current_version"),
                    values.get("latest_version"),
                    values.get("updated"),
                ], divider=True)
        
        click.echo(item_table)



from __future__ import annotations

from typing import Literal

import ckan.plugins.toolkit as tk

SELFINFO_REDIS_PREFIX = 'ckan.selfinfo.redis_prefix_key'
SELFINFO_ERRORS_LIMIT = 'ckan.selfinfo.errors_limit'
SELFINFO_REPOS_PATH = 'ckan.selfinfo.ckan_repos_path'
SELFINFO_REPOS = 'ckan.selfinfo.ckan_repos'
SELFINFO_PARTITIONS_PATH = 'ckan.selfinfo.partitions'
SELFINFO_REDIS_SUFFIX: Literal["_selfinfo"] = "_selfinfo"
STORE_TIME: float = 604800.0 # one week
# STORE_TIME: float = 1.0
PYPI_URL: Literal["https://pypi.org/pypi/"] = "https://pypi.org/pypi/"


def selfinfo_get_redis_prefix():
    prefix = tk.config.get(SELFINFO_REDIS_PREFIX, '')
    return prefix + "_" if prefix else prefix


def selfinfo_get_errors_limit():
    return int(tk.config.get(SELFINFO_ERRORS_LIMIT, 20))


def selfinfo_get_repos_path():
    return tk.config.get(SELFINFO_REPOS_PATH)


def selfinfo_get_repos():
    return tk.config.get(SELFINFO_REPOS, '')

def selfinfo_get_partitions():
    return tk.config.get(SELFINFO_PARTITIONS_PATH, '/')



from __future__ import annotations

from typing import Any

import ckanext.selfinfo.utils as selfinfoutils


class TestUTILS:
    def test_get_python_modules_info(self):    
        python_moduls_info: dict[str, Any] = selfinfoutils.get_python_modules_info()
        
        assert type(python_moduls_info) == dict
        
        assert len(python_moduls_info.keys()) == 3
    
    def test_get_lib_data(self):
        lib_data: dict[str, Any] | None = selfinfoutils.get_lib_data("ckan")
        
        assert type(lib_data) == dict
        
        assert lib_data["info"]["name"] == "ckan"
        
    def test_get_lib_latest_version(self):
        lib_data_latest_version: str | None = selfinfoutils.get_lib_latest_version("ckan")
        
        assert type(lib_data_latest_version) == str

    def test_get_ram_usage(self):
        ram_usage: dict[str, Any] = selfinfoutils.get_ram_usage()
        
        assert type(ram_usage) == dict
        
        assert len(ram_usage.keys()) == 2
        
    def test_get_platform_info(self):
        platform_info: dict[str, Any] = selfinfoutils.get_platform_info()
        
        assert type(platform_info) == dict
        
        assert len(platform_info.keys()) == 2



from __future__ import annotations

from typing import Any


from ckan import types
import ckan.plugins.toolkit as tk
import ckan.plugins as p

import ckanext.selfinfo.utils as selfutils
from ckanext.selfinfo.interfaces import ISelfinfo


@tk.side_effect_free
def get_selfinfo(
    context: types.Context,
    data_dict: dict[str, Any],
) -> dict[str, Any]:
    
    tk.check_access("sysadmin", context, data_dict)
    
    platform_info: dict[str, Any] = selfutils.get_platform_info()
    ram_usage: dict[str, Any] = selfutils.get_ram_usage()
    disk_usage: list[dict[str, Any]] = selfutils.get_disk_usage()
    groups: dict[str, Any] = selfutils.get_python_modules_info(
        force_reset=data_dict.get("force-reset", False),
    )
    freeze = selfutils.get_freeze()
    git_info = selfutils.gather_git_info()
    errors = selfutils.retrieve_errors()
    actions = selfutils.ckan_actions()
    blueprints = selfutils.ckan_bluprints()
    helpers = selfutils.ckan_helpers()

    data = {
        "groups": groups,
        "platform_info": platform_info,
        "ram_usage": ram_usage,
        "disk_usage": disk_usage,
        "git_info": git_info,
        "freeze": freeze,
        "errors": errors,
        "actions": actions,
        "blueprints": blueprints,
        "helpers": helpers,
    }

    # data modification
    for item in p.PluginImplementations(ISelfinfo):
        item.selfinfo_after_prepared(data)

    return data



from __future__ import annotations

import os
from typing import Any, Mapping
import requests
import psutil
from psutil._common import bytes2human
import platform
import git
from datetime import datetime
import importlib_metadata as imetadata
import logging
import json
import distro
import inspect

from ckan.lib.redis import connect_to_redis, Redis
import ckan.plugins.toolkit as tk
from .config import (
    selfinfo_get_redis_prefix,
    selfinfo_get_repos_path,
    selfinfo_get_partitions,
    SELFINFO_REDIS_SUFFIX,
    STORE_TIME,
    PYPI_URL,
)

log = logging.getLogger(__name__)


def get_redis_key(name):
    return selfinfo_get_redis_prefix() + name + SELFINFO_REDIS_SUFFIX


def get_python_modules_info(force_reset: bool=False) -> dict[str, Any]:
    redis: Redis = connect_to_redis()    
    now: float = datetime.utcnow().timestamp()
    
    groups: dict[str, Any] = {"ckan": {}, "ckanext": {}, "other": {}}
    pdistribs: Mapping[str, Any] = imetadata.packages_distributions()
    modules: dict[str, Any] = {
        p.name: p.version for p in imetadata.distributions()}

    for i, p in pdistribs.items():
            for module in p:
                group: str = i if i in groups else "other"

                if module in module and not module in groups[group]:
                    redis_key: str = get_redis_key(module)
                    data: Mapping[str, Any] = {
                        "name": module,
                        "current_version": modules.get(module, 'unknown'),
                        "updated": now,
                    }
                    if not redis.hgetall(redis_key):
                        data["latest_version"] = get_lib_latest_version(module)
                        redis.hset(redis_key, mapping=data)

                    if (now - float(redis.hget(redis_key, "updated").decode("utf-8"))) > STORE_TIME or force_reset:
                        data["latest_version"] = get_lib_latest_version(module)
                        for key in data:
                            if data[key] != redis.hget(redis_key, key):
                                redis.hset(redis_key, key=key, value=data[key])

                    groups[group][module] = {k.decode("utf-8"): v.decode("utf-8") for k, v in redis.hgetall(redis_key).items()}
    
                    groups[group][module]["updated"] = datetime.fromtimestamp(float(groups[group][module]["updated"]))

    groups["ckanext"] = dict(sorted(groups["ckanext"].items()))
    groups["other"] = dict(sorted(groups["other"].items()))

    return groups


def get_freeze():
    try:
        from pip._internal.operations import freeze
    except ImportError: # pip < 10.0
        from pip.operations import freeze
    pkgs = freeze.freeze()
    pkgs = list(pkgs)
    pkgs_string = "\n".join(list(pkgs))
    return {
        "modules": pkgs,
        "modules_html": f"""{pkgs_string}""",
    }


def get_lib_data(lib):
    req = requests.get(PYPI_URL + lib + '/json', headers={
        "Content-Type": "application/json"
    })

    if req.status_code == 200:
        return req.json()
    return None


def get_lib_latest_version(lib):
    data = get_lib_data(lib)
    
    if data and data.get('info'):
        return data['info'].get('version', 'unknown')
    return 'unknown'


def get_ram_usage() -> dict[str, Any]:
    memory = psutil.virtual_memory()
    return {
        "precent_usage": memory.percent,
        "used_ram": bytes2human(memory.used, format="%(value).1f"),
        "total_ram": bytes2human(memory.total, format="%(value).1f")
    }


def get_disk_usage():
    paths = selfinfo_get_partitions()
    results = []
    for path in paths.split(','):
        try:
            usage = psutil.disk_usage(path.strip())
            if usage:
                results.append(
                    {
                        "path": path,
                        "precent_usage":  usage.percent,
                        "total_disk": bytes2human(usage.total, format="%(value).1f%(symbol)s"),
                        "free_space": bytes2human(usage.free, format="%(value).1f%(symbol)s")
                    }
                )
        except OSError:
            log.exception(f"Path '{path}' does not exists.")
    return results


def get_platform_info() -> dict[str, Any]:
    return {
        "distro": distro.name() + " " + distro.version(),
        "python_version": platform.python_version(),
        "platform": platform.platform(),
    }


def gather_git_info():
    ckan_repos_path = selfinfo_get_repos_path()
    git_info = {
        "repos_info": [],
        "access_errors": {}
    }
    if ckan_repos_path:
        ckan_repos = tk.config.get('ckan.selfinfo.ckan_repos', '')
        list_repos = ckan_repos.strip().split(" ") if \
            ckan_repos else [
                name for name in os.listdir(
                    ckan_repos_path) if os.path.isdir(
                        os.path.join(ckan_repos_path, name)) and not name.startswith('.')
        ]

        repos: dict[str, git.Repo | None] = {
            repo: get_git_repo(ckan_repos_path + '/' + repo) for repo in list_repos if repo}

        for name, repo in repos.items():
            if not repo:
                continue
            try:
                commit, branch = repo.head.object.name_rev.strip().split(" ")
                short_sha: str = repo.git.rev_parse(commit, short=True)
                on = 'branch'

                if repo.head.is_detached and branch.startswith("remotes/"):
                    branch = short_sha
                    on = 'commit'
                elif repo.head.is_detached and branch.startswith("tags/"):
                    on = 'tag'
                elif repo.head.is_detached and (
                    not branch.startswith("tags/") and not branch.startswith("remotes/")):
                    branch = short_sha
                    on = 'commit'

                git_info['repos_info'].append({
                    "name": name,
                    "head": branch,
                    "commit": short_sha,
                    "on": on,
                    "remotes": [
                        {
                            "name": remote.name,
                            "url": remote.url,
                            } for remote in repo.remotes
                        ]
                })
            except ValueError as e:
                git_info['access_errors'][name] = e
    return git_info


def get_git_repo(path):
    repo = None
    try:
        repo = git.Repo(path)
    except git.exc.InvalidGitRepositoryError as e:
        pass

    return repo


def retrieve_errors():
    redis: Redis = connect_to_redis()
    key = get_redis_key('errors')
    if not redis.exists(key):
        redis.set(key, json.dumps([]))
    return json.loads(redis.get(key))


def ckan_actions():
    from ckan.logic import _actions
    data = []
    for n, f in _actions.items():
        data.append({
            "func_name": n,
            "docstring": inspect.getdoc(f),
        })

    return data


def ckan_bluprints():
    from flask import current_app
    app = current_app
    data = {}

    for name, blueprint in app.blueprints.items():
        data[name] = []
        for rule in current_app.url_map.iter_rules():
            if rule.endpoint.startswith(f"{name}."):
                view_func = current_app.view_functions[rule.endpoint]
                # signature = inspect.signature(view_func)

                data[name].append({
                    'path': rule.rule,
                    'methods': rule.methods,
                    'route': rule.endpoint,
                    'route_func': view_func.__name__,
                })

    return data


def ckan_helpers():
    from ckan.lib.helpers import helper_functions
    data = []
    for n, f in helper_functions.items():
        data.append({
            "func_name": n,
            "docstring": inspect.getdoc(f),
            "defined": inspect.getsourcefile(f)
        })
    return data



from __future__ import annotations

from ckan.plugins import Interface


class ISelfinfo(Interface):
    """Implement custom Selfinfo response modification."""

    def selfinfo_after_prepared(self, data):
        """Return selinfo data.

        :returns: dictonary
        :rtype: dict

        """

        return data



from __future__ import annotations

from typing import Any

from . import get
from . import update


def get_actions():
    actions: dict[str, Any] = {
        "get_selfinfo": get.get_selfinfo,
        "update_last_module_check": update.update_last_module_check,
    }

    return actions



import click

from . import main

def get_commands():
    return [selfinfo]


@click.group()
def selfinfo():
    """ckanext-selfinfo management commands."""
    
selfinfo.command(main.update_module_info)
selfinfo.command(main.get_selfinfo)



from __future__ import annotations

import logging
import json
from datetime import datetime

from ckan.lib.redis import connect_to_redis, Redis

from .config import selfinfo_get_errors_limit
from .utils import get_redis_key


class SelfinfoErrorHandler(logging.Handler):
    """Custom handler to store exceptions."""
    def emit(self, record):
        if record.levelno >= logging.ERROR:
            redis: Redis = connect_to_redis()    
            now: float = datetime.utcnow().timestamp()
            log_message = self.format(record)
            redis_key = get_redis_key('errors')

            if not redis.exists(redis_key):
                redis.set(redis_key, json.dumps([]))

            erorrs: list = json.loads(redis.get(redis_key))
            errors_limit = selfinfo_get_errors_limit()
            if len(erorrs) >= errors_limit:
                start_key = len(erorrs) - errors_limit + 1
                erorrs = erorrs[start_key:]

            erorrs.append(log_message)
            redis.set(redis_key, json.dumps(erorrs))



from __future__ import annotations

import json
from typing import Any, cast
from flask import Blueprint
from flask.views import MethodView

from ckan import types
import ckan.model as model
import ckan.plugins.toolkit as tk
from ckan.lib.redis import connect_to_redis, Redis
from ckanext.selfinfo.utils import get_redis_key


selfinfo = Blueprint("selfinfo", __name__, url_prefix="/ckan-admin")


class SelfinfoView(MethodView):
    def get(self):
        try:
            context: types.Context = cast(types.Context, {
                "model": model,
                "user": tk.current_user.name,
                "auth_user_obj": tk.current_user
            })

            tk.check_access(u'sysadmin', context)
        except tk.NotAuthorized:
            tk.abort(404)
        
        args = tk.request.args

        if args.get('drop_errors') and tk.asbool(args['drop_errors']):
            redis: Redis = connect_to_redis()
            key = get_redis_key('errors')
            redis.set(key, json.dumps([]))

            return tk.redirect_to("selfinfo.index")

        data: dict[str, Any] = tk.get_action("get_selfinfo")({}, {})
        status_show: dict[str, Any] = tk.get_action("status_show")({}, {})

        return tk.render("selfinfo/index.html", {
            "data": data,
            "status_show": status_show,
            }
        )


selfinfo.add_url_rule(
    "/selfinfo",
    view_func=SelfinfoView.as_view("index"),
)



from __future__ import annotations

from typing import Any
import pytest
import platform
import os

from ckan import model
import ckan.plugins.toolkit as tk
from ckan.tests.helpers import call_action
import ckan.tests.factories as factories

current_path: list[str] = os.getcwd().split("/")
current_path.pop()
updated_path: str = "/".join(current_path)


@pytest.mark.ckan_config("ckan.plugins", "selfinfo")
@pytest.mark.usefixtures("with_plugins", "clean_db")
@pytest.mark.ckan_config("ckan.selfinfo.ckan_repos_path", updated_path)
@pytest.mark.ckan_config("ckan.selfinfo.ckan_repos", "ckan ckanext-selfinfo")
class TestGET:
    def test_get_selfinfo(self):
        user = factories.User()
        sysadmin = factories.Sysadmin()
        context: dict[str, Any] = {
            "model": model,
            "user": user["name"],
            "ignore_auth": False,
        }

        with pytest.raises(tk.NotAuthorized):
            call_action("get_selfinfo", context=context)

        context["user"] = sysadmin['name']
        
        selfinfo: dict[str, Any] = tk.get_action("get_selfinfo")(context, {})

        assert type(selfinfo) == dict
        
        assert len(selfinfo.keys()) == 4
        
        assert selfinfo["platform_info"]["python_version"] == platform.python_version()
